@@ -1,17 +1,18 @@
 
 # The R Commander and command logger
 
-# last modified 19 December 2008 by J. Fox
+# last modified 24 January 2010 by J. Fox
 #   slight changes 12 Aug 04 by Ph. Grosjean
 #   changes 21 June 2007 by Erich Neuwirth for Excel support (marked EN)
-# last modified 17 December 2008 by Richard Heiberger  ##rmh
+#   modified 17 December 2008 by Richard Heiberger  ##rmh
 
 Commander <- function(){
-	RcmdrVersion <- "1.4-6"
+	RcmdrVersion <- "1.5-6"
 	##    DESCRIPTION <- readLines(file.path(.find.package("Rcmdr"), "DESCRIPTION")[1])
 	##    RcmdrVersion <- trim.blanks(sub("^Version:", "",
 	##        grep("^Version:", D, value=TRUE)))
 	putRcmdr("quotes", options(useFancyQuotes=FALSE))
+	putRcmdr("messageNumber", 0)
 	# the following test suggested by Richard Heiberger
 	if ("RcmdrEnv" %in% search() &&
 		exists("commanderWindow", "RcmdrEnv") &&
@@ -26,6 +27,7 @@ Commander <- function(){
 		else opt
 	}
 	current <- options("Rcmdr")[[1]]
+	setOption("number.messages", TRUE)
 	etc <- setOption("etc", file.path(.path.package(package="Rcmdr")[1], "etc"))
 	etcMenus <- setOption("etcMenus", etc)
 	putRcmdr("etcMenus", etcMenus)
@@ -159,6 +161,7 @@ Commander <- function(){
 				else 20, global=FALSE))
 	putRcmdr("saveOptions", options(warn=1, contrasts=getRcmdr("default.contrasts"), width=as.numeric(log.width),
 			na.action="na.exclude", graphics.record=TRUE))
+	setOption("ask.to.exit", TRUE)
 	setOption("ask.on.exit", TRUE)
 	setOption("double.click", FALSE)
 	setOption("sort.names", TRUE)
@@ -169,15 +172,20 @@ Commander <- function(){
 	setOption("output.text.color", "darkblue")
 	setOption("error.text.color", "red")
 	setOption("warning.text.color", "darkgreen")
+	setOption("prefixes", c("Rcmdr> ", "Rcmdr+ ", "RcmdrMsg: ", "RcmdrMsg+ "))
 	setOption("multiple.select.mode", "extended")
 	setOption("suppress.X11.warnings",
-		interactive() && .Platform$GUI == "X11" && getRversion() < "2.4.0") # to address problem in Linux
+		interactive() && .Platform$GUI == "X11") # to address problem in X11 (Linux or Mac OS X)
+#		interactive() && .Platform$GUI == "X11" && getRversion() < "2.4.0")
 	setOption("showData.threshold", 100)
 	setOption("retain.messages", TRUE)
 	setOption("crisp.dialogs",  TRUE)
 	setOption("length.output.stack", 10)
+	setOption("length.command.stack", 10)
 	putRcmdr("outputStack", as.list(rep(NA, getRcmdr("length.output.stack"))))
+	putRcmdr("commandStack", as.list(rep(NA, getRcmdr("length.command.stack"))))
 	setOption("variable.list.height", 4)
+	setOption("variable.list.width", c(20, Inf))
 	if (getRcmdr("suppress.X11.warnings")) {
 		putRcmdr("messages.connection", file(open = "w+"))
 		sink(getRcmdr("messages.connection"), type="message")
@@ -309,17 +317,18 @@ Commander <- function(){
 		tkfocus(CommanderWindow())
 	}
 	onView <- function(){
+		if (packageAvailable("relimp")) Library("relimp")
 		if (activeDataSet() == FALSE) {
 			tkfocus(CommanderWindow())
 			return()
 		}
+		suppress <- if(getRcmdr("suppress.X11.warnings")) ", suppress.X11.warnings=FALSE" else ""
 		view.height <- max(as.numeric(output.height) + as.numeric(log.height), 10)
 		ncols <- ncol(get(ActiveDataSet()))
 #        ncols <- eval(parse(text=paste("ncol(", ActiveDataSet(), ")")))
 		command <- if (packageAvailable("relimp") && ncols <= getRcmdr("showData.threshold")){
-				require("relimp")
 				paste("showData(", ActiveDataSet(), ", placement='-20+200', font=getRcmdr('logFont'), maxwidth=",
-					log.width, ", maxheight=", view.height, ")", sep="")
+					log.width, ", maxheight=", view.height, suppress, ")", sep="")
 			}
 			else paste("View(", ActiveDataSet(), ")", sep="")
 		logger(command)
@@ -352,7 +361,7 @@ Commander <- function(){
 			while (nchar(lines[iline])==0) iline <- iline + 1
 			if (iline > nlines) break
 			current.line <- lines[iline]
-			if (.console.output) cat(paste("\nRcmdr> ", current.line,"\n", sep=""))
+			if (.console.output) cat(paste("\n", getRcmdr("prefixes")[1], current.line,"\n", sep=""))
 			else{
 				tkinsert(.output, "end", paste("\n> ", current.line,"\n", sep="")) ### end of changed
 				tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend")
@@ -361,7 +370,7 @@ Commander <- function(){
 			jline <- iline + 1
 			while (jline <= nlines){
 				if (class(try(parse(text=current.line),silent=TRUE))!="try-error") break
-				if (.console.output)cat(paste("Rcmdr+ ", lines[jline],"\n", sep=""))
+				if (.console.output)cat(paste(getRcmdr("prefixes")[2], lines[jline],"\n", sep=""))
 				else{
 					tkinsert(.output, "end", paste("+ ", lines[jline],"\n", sep=""))
 					tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend")
@@ -429,12 +438,12 @@ Commander <- function(){
 				if (Menus[m, 3] == "command"){
 					position[Menus[m, 2]] <- position[Menus[m, 2]] + 1
 					if (Menus[m, 6] == "")
-						tkadd(get(Menus[m, 2]), "command", label=gettextRcmdr(Menus[m, 4]),
+						tkadd(get(Menus[m, 2]), "command", label=gettextMenus(Menus[m, 4]),
 							command=get(Menus[m, 5]))
 #                        tkadd(eval(parse(text=Menus[m, 2])),"command", label=gettextRcmdr(Menus[m, 4]),
 #                            command=eval(parse(text=Menus[m, 5])))
 					else {
-						tkadd(get(Menus[m, 2]), "command", label=gettextRcmdr(Menus[m, 4]),
+						tkadd(get(Menus[m, 2]), "command", label=gettextMenus(Menus[m, 4]),
 							command=get(Menus[m, 5]), state="disabled")
 #                        tkadd(eval(parse(text=Menus[m, 2])),"command", label=gettextRcmdr(Menus[m, 4]),
 #                            command=eval(parse(text=Menus[m, 5])),  state="disabled")
@@ -445,7 +454,7 @@ Commander <- function(){
 					}
 				}
 				else if (Menus[m, 3] == "cascade")
-					tkadd(get(Menus[m, 2]), "cascade", label=gettextRcmdr(Menus[m, 4]),
+					tkadd(get(Menus[m, 2]), "cascade", label=gettextMenus(Menus[m, 4]),
 						menu=get(Menus[m, 5]))
 #                    tkadd(eval(parse(text=Menus[m, 2])),"cascade", label=gettextRcmdr(Menus[m, 4]),
 #                        menu=eval(parse(text=Menus[m, 5])))
@@ -568,10 +577,11 @@ Commander <- function(){
 	tkbind(.commander, "<Control-X>", onCut)
 	tkbind(.commander, "<Control-c>", onCopy)
 	tkbind(.commander, "<Control-C>", onCopy)
-	if (.Platform$OS.type != "windows"){
-		tkbind(.commander, "<Control-v>", onPaste)
-		tkbind(.commander, "<Control-V>", onPaste)
-	}
+#	if (.Platform$OS.type != "windows"){
+#		tkbind(.commander, "<Control-v>", onPaste)
+#		tkbind(.commander, "<Control-V>", onPaste)
+#		tkbind(.commander, "<Alt-BackSpace>", onUndo)
+#	}
 	tkbind(.commander, "<Control-r>", onSubmit)
 	tkbind(.commander, "<Control-R>", onSubmit)
 	tkbind(.commander, "<Control-Tab>", onSubmit)
@@ -595,14 +605,18 @@ Commander <- function(){
 	}
 	##    if (!packageAvailable("rgl")) Message(gettextRcmdr("The rgl package is absent; 3D plots are unavailable."), type="warning")
 	Message(paste(gettextRcmdr("R Commander Version "), getRcmdr("RcmdrVersion"), ": ", date(), sep=""))
+	if (.Platform$GUI == "Rgui"  && ismdi()) Message(gettextRcmdr(
+		"The Windows version of the R Commander works best under RGui\nwith the single-document interface (SDI); see ?Commander."),
+		type="warning")
 }
 
 
 # the following function modified 24 July 07 by Richard Heiberger
 #  and subsequently by J. Fox 26 July 07
-# last modified 3 December 08 by J. Fox
+# last modified 10 January 2010 by J. Fox
 
 logger <- function(command){
+	pushCommand(command)
 	if (is.SciViews()) return(svlogger(command))    # +PhG
 	.log <- LogWindow()
 	.output <- OutputWindow()
@@ -616,14 +630,14 @@ logger <- function(command){
 	lines <- strsplit(command, "\n")[[1]]
 	tkinsert(.output, "end", "\n")
 	if (getRcmdr("console.output")) {
-		for (line in seq(along=lines)) {
-			prompt <- ifelse (line==1, "\nRcmdr>", "\nRcmdr+")
+		for (line in seq(along.with=lines)) {
+			prompt <- ifelse (line==1, paste("\n", getRcmdr("prefixes")[1], sep=""), paste("\n", getRcmdr("prefixes")[2], sep=""))
 			cat(paste(prompt, lines[line]))  ##rmh
 		}
 		cat("\n")                          ##rmh
 	}
 	else {
-		for (line in  seq(along=lines)) {
+		for (line in  seq(along.with=lines)) {
 			prompt <- ifelse(line==1, "> ", "+ ")
 			tkinsert(.output, "end", paste(prompt, lines[line], "\n", sep=""))
 			tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend")
@@ -680,7 +694,7 @@ doItAndPrint <- function(command, log=TRUE) {
 			if (!.console.output) sink(type="output") # if .console.output, output connection already closed
 			close(output.connection)
 		}, add=TRUE)
-	if (log) logger(command)
+	if (log) logger(command) else pushCommand(command)
 	result <- try(parse(text=paste(command)), silent=TRUE)
 	if (class(result)[1] == "try-error"){
 		Message(message=paste(strsplit(result, ":")[[1]][2]), type="error")
@@ -730,8 +744,9 @@ doItAndPrint <- function(command, log=TRUE) {
 
 checkWarnings <- function(messages){
 	if (getRcmdr("suppress.X11.warnings")){
-		X11.warning <- grep("^Warning\\: X11 protocol error\\: BadWindow \\(invalid Window parameter\\)",
-			messages)
+#		X11.warning <- grep("^Warning\\: X11 protocol error\\: BadWindow \\(invalid Window parameter\\)",
+#			messages)
+		X11.warning <- grep("X11 protocol error|Warning in structure", messages)
 		if (length(X11.warning) > 0){
 			messages <- messages[-X11.warning]
 		}
@@ -771,9 +786,10 @@ Message <- function(message, type=c("note", "error", "warning")){
 	type <- match.arg(type)
 	if (type != "note") tkbell()
 	if (getRcmdr("retain.messages")) {
-		if (!missing(message)) tkinsert(.message, "end", "\n")
-		else if (!is.null(getRcmdr("last.message"))) {
-			tkinsert(.message, "end", "\n\n")
+#		if (!missing(message)) tkinsert(.message, "end", "\n")
+#		else if (!is.null(getRcmdr("last.message"))) {
+		if (missing(message) && !is.null(getRcmdr("last.message"))) {
+#			tkinsert(.message, "end", "\n\n")
 			putRcmdr("last.message", NULL)
 			tkyview.moveto(.message, 1.0)
 		}
@@ -793,6 +809,11 @@ Message <- function(message, type=c("note", "error", "warning")){
 	}
 	putRcmdr("last.message", type)
 	message <- paste(prefix, ": ", message, sep="")
+	if (getRcmdr("retain.messages") && getRcmdr("number.messages")) {
+		messageNumber <- getRcmdr("messageNumber") + 1
+		putRcmdr("messageNumber", messageNumber)
+		message <- paste("[", messageNumber, "] ", message, sep="")
+	}
 	######### added by EN #####################
 	if (RExcelSupported())
 		putRExcel(".rexcel.last.message",message)
@@ -802,8 +823,8 @@ Message <- function(message, type=c("note", "error", "warning")){
 	######### added by rmh #####################                   ##rmh
 	if (console.output) {                                        ##rmh & J. Fox
 		if (sink.number() != 0) sink()							## fixed by J. Fox
-		for (jline in seq(along=lines)) {                            ##rmh
-			Header <- if (jline==1) "RcmdrMsg: " else "RcmdrMsg+ "     ##rmh
+		for (jline in seq(along.with=lines)) {                            ##rmh
+			Header <- if (jline==1) getRcmdr("prefixes")[3] else getRcmdr("prefixes")[4]     ##rmh
 			cat(paste(Header, lines[jline], "\n", sep=""))             ##rmh
 		}                                                            ##rmh
 	}                                                              ##rmh
@@ -841,3 +862,16 @@ popOutput <- function(){
 	putRcmdr("outputStack", c(stack[-1], NA))
 	lastOutput
 }
+
+pushCommand <- function(element) {
+	stack <- getRcmdr("commandStack")
+	stack <- c(list(element), stack[-getRcmdr("length.command.stack")])
+	putRcmdr("commandStack", stack)
+}
+
+popCommand <- function(){
+	stack <- getRcmdr("commandStack")
+	lastCommand <- stack[[1]]
+	putRcmdr("commandStack", c(stack[-1], NA))
+	lastCommand
+}
